﻿/****************************************************
 * SpectrogramWithFft
 * 
 * [purpose]
 *              本プログラムの開発目的は、WAVEファイル読み込みクラスのサンプルを提供することです。
 *              
 *              裏の目的は、スペクトログラムの高速表示の方法を探る・・・です。
 *              作ってみて分かったのですが、メモリの制約がうるさいですね。
 *              2万とちょっとのフレームしかデータを保持することができませんでした。
 *              これは4096 pointなら30 min（@44.1 kHz sampling）程度です。
 *              スペクトログラムを簡単にみることができるAdobe Auditionが良くできていることが分かりました。
 *              
 *              もしAdobe Auditionのまねをするなら、周波数方向の表示分解能を固定してしまう必要があると思います。
 *              その上で時間幅ごとにキャッシュを作成してバイナリでファイルを保存し、アドレスで時間を指定できるようにしておきます。
 *              表示に必要な部分を（新たに必要な分だけ）随時高速読み出しすると割と早い表示ができると思います。
 *              もし、リアルタイム表示するならFFTresultのトコロテン式バッファクラスと、それを内蔵するimgデータ作成クラスが必要だと思います。
 * [abst.]
 *              WAVEファイルを読み込み、FFTを実施してスペクトログラムを作ります。
 *              デフォルトでは、音声ファイルの始めの512フレームが使われます。
 *              FFTにかけるサンプル数によって、時間分解能と周波数分解能のトレードオフ関係が目で確認できるので分かり易いです。
 *              時間分解能の面ではウェーブレットの方が使い易いんだろうなぁと思いました。
 * [programmer]
 *              K. Morishita (Kumamoto-Univ. @ 2012.11)
 * [history]
 *              2012/11/2   FFTとWAVEのサンプルとして作成
 * **************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Threading.Tasks;

using Signal;
using Sound;
using Sound.WAVE;
using Signal.FrequencyAnalysis;
using Graphic.ColorClass;

namespace SpectrogramWithFft
{
    public partial class Form1 : Form
    {
        /* メンバ変数 *******************************************/
        /// <summary>
        /// WAVEファイルを読み込むオブジェクト
        /// </summary>
        Sound.WAVE.WaveReader reader = new Sound.WAVE.WaveReader();
        /* プロパティ *******************************************/
        /* メソッド *******************************************/
        public Form1()
        {
            InitializeComponent();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog hoge = new OpenFileDialog();
            hoge.FileName = "hoge.wave";                                 // ファイルを開く前に初期設定
            hoge.Filter = "Wave Files (*.wav, *.wave)|*.wav;*.wave";
            if (hoge.ShowDialog() == DialogResult.OK)
            {
                this.reader.Open(hoge);

                /*
                // 並列計算を可能にして、高速化できると思ったら、メモリが足らんかった。
                // しかも、遅い気がする。
                List<double[]> listSound = new List<double[]>();
                while (this.reader.IsOpen && this.reader.ReadLimit == false)
                {
                    var data = this.reader.Read(1024);
                    double[] sound;
                    if (data.UsableCH == MusicData.UsableChannel.Left)
                        sound = data.GetData(MusicData.Channel.Left);
                    else
                        sound = data.GetData(MusicData.Channel.Right);

                    listSound.Add(sound);
                }

                FFTresult[] listBuff = new FFTresult[listSound.Count];
                Parallel.For(0, listSound.Count, i =>
                {
                    var dft = new DFT(listSound[i].Length, Window.WindowKind.Hanning);
                    FFTresult result = dft.FFT(listSound[i], (long)this.reader.SamplingRate);
                    listBuff[i] = result;
                });
                */

                if(this.reader.IsOpen && this.reader.ReadLimit == false)                // ファイルを開くことができているかチェック
                {
                    // 指定フレーム数だけ読み出す
                    int i = 0;
                    List<FFTresult> listBuff = new List<FFTresult>();
                    while (i < 512)                                                     // 数を絞らないと、メモリが足らなくなる
                    {
                        var data = this.reader.Read(1024);                              // 指定ポイント数読み出す
                        double[] sound;
                        if (data.UsableCH == MusicData.UsableChannel.Left)              // データの入っている方を格納する
                            sound = data.GetData(MusicData.Channel.Left);
                        else
                            sound = data.GetData(MusicData.Channel.Right);

                        var dft = new DFT(sound.Length, Window.WindowKind.Hanning);
                        FFTresult result = dft.FFT(sound, (long)this.reader.SamplingRate);
                        listBuff.Add(result);
                        
                        i++;
                    }

                    // 画像化（画像としてファイルへ出力するなら、Graphic.BmpGeneratorを使った方が良い）
                    double max = double.MinValue;                                       // 中央値の最大を探す
                    for (int j = 0; j < listBuff.Count; j++)
                        if (max < listBuff[j].MaxAbs) max = listBuff[j].MedianAbs;
                    int height = 512;                                                   // 画像の高さを設定
                    Bitmap img = new Bitmap(listBuff.Count, height); 
                    Graphic.ColorClass.Colormap cmap = new Colormap();
                    for (int k = 0; k < height; k++)
                    {
                        for (int j = 0; j < listBuff.Count; j++)
                        {
                            //色を決める
                            int index = listBuff[j].Length * k / height;
                            double power = 3.0;                                         // 表示幅を決定する。3.0だと、中央値の1000倍～0.001倍までを表示させる。2.0だと100～0.01倍の範囲。
                            var value = (Math.Log10(listBuff[j].GetData(index).data.Abs / max) + power) / (power * 2.0);
                            if (value > 1.0) value = 1.0;
                            if (value < 0.0) value = 0.0;
                            //Console.WriteLine(value);
                            Color c = cmap.GetColor(value);
                            //1つのピクセルの色を変える
                            img.SetPixel(j, height - k - 1, c);
                        }
                    }
                    //作成した画像を表示する
                    this.pictureBox1.Image = img;
                }
            }
            
        }
    }
}
